/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  main.c
  
  ***************************************/


#include <sys/wait.h>
#include <sys/signal.h>
#include <errno.h>

#include "ftp-proxy.h"
#include "common.h"

int run_daemon(void);
int accept_connection(struct sockaddr_in client);

void daemonify(void);
void write_pidfile(void);
static RETSIGTYPE signal_handle(int signo);
static int listen_sockfd, noforks=0;

static int reread_flag=FALSE;

int main(int argc, char *argv[])
{
	struct sockaddr_in listen_address;
	
	process_cmdline(argc, argv);

	if( read_config()!=0 ){
		fprintf(stderr, "Error reading configuration file\n");
		exit(1);
	}

	if(config_maxforks()!=0) daemonify();

	if(config_transdata()) {
		sufd=transdata_setup();
		if(sufd==-1) config_transdatafail();
	}

	write_pidfile();

	listen_address=config_getaddress();

	if( (listen_sockfd=listen_on_socket(&listen_address, CTRL))==-1) exit(1);

	config_openlog();

	write_log(INFO, "Listening on %s:%d",
		  addr2name(listen_address.sin_addr),
		  ntohs(listen_address.sin_port));

	config_bindtodevice(listen_sockfd);
	config_chroot();
	config_droppriv();

	signal(SIGCHLD,signal_handle);
	signal(SIGINT,signal_handle);
	signal(SIGTERM,signal_handle);
	signal(SIGHUP,signal_handle);
	signal(SIGPIPE,SIG_IGN);

	if(config_cacheinit()==-1){
		fprintf(stderr, "Unable to initialise cache. Exiting\n");
		exit(1);
	}
	run_daemon();

	return(0);
}
   
int run_daemon()
{
	struct sockaddr_in client_address;
	int len = sizeof(client_address);
	int fd;

	do{
		fd = accept( listen_sockfd, (struct sockaddr *)&client_address, &len);
		if(fd < 0 ){
			if(errno==EINTR) continue;
			perror("accept");
			quit(SOCKET);
		}
		if(reread_flag){
			reread_flag=FALSE;
			reread_config(&listen_sockfd);
		}

		if (!accept_connection(client_address)){
			close(fd);
		} else {
			pid_t pid;
			
			if(config_maxforks()==0){
				close(listen_sockfd);
				init_session(fd, client_address);
				exit(0);
			}

			pid = fork();
			switch(pid)
			{
			case -1:
				perror("fork failed");
				close(fd);
				break;
			case 0:
				signal(SIGHUP, SIG_IGN);
				init_session(fd, client_address);
				exit(0);
			default:
				close(fd);
				noforks++;
			}
		}
	} while(TRUE);

	freopen("/dev/null", "w", stdin );
	freopen("/dev/null", "w", stdout);
}

int accept_connection(struct sockaddr_in client)
{

	if(noforks>=config_maxforks() && config_maxforks()!=0) {
		write_log(ERROR, "Connect from %s refused: Too many connections",
			  inet_ntoa(client.sin_addr));
		return(FALSE);
	}

	if(config_bouncedefend() && ntohs(client.sin_port)==20) {
		write_log(ATTACK, "Connect from %s refused: Comes from ftp-data port",
			  inet_ntoa(client.sin_addr));
		return(FALSE);
	}

	return(TRUE);
}

/* ------------------------------------------------------------- **
** Name says it worse than I ever could!
** ------------------------------------------------------------- */
void daemonify()
{
	switch (fork()) {
	case -1:
		write_log(ERROR, "can't fork daemon");
		exit(-1);
		break;
	case 0:
		/******** child ********/
		break;
	default:
		/******** parent ********/
		exit(0);
	}

	debug("Forked to background\n");
}

void write_pidfile(void)
{
	FILE *fp;
	pid_t pid;

	pid=getpid();

	if(config_pidfile()){
		fp=fopen(config_pidfile(), "w");
		if(fp!=NULL) {
			fprintf(fp, "%d\n", pid);
			fclose(fp);
		}
	}
}

static RETSIGTYPE signal_handle(int signo)
{
	switch(signo){
	case SIGCHLD :
		while(waitpid(-1, (int *) 0, WNOHANG)>0)
			noforks--;
#ifdef TRANS_DATA
		if(noforks==0 && config_transdata()) {
			mangle_req req;
			req.type=FLUSH;
			write(sufd, &req, sizeof(req));
		}
#endif
		signal(signo, signal_handle);
		break;
	case SIGINT :
	case SIGTERM :
		close(listen_sockfd);
		exit(0);
	case SIGHUP : /*Reread config file. Not safe from signal handler.*/
		reread_flag=TRUE;

	}
}
